@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
-# pylint: disable-msg=C0301,W0613,W0612
+# pylint: disable-msg=C0301,W0613,W0612,W0212,W0511,R0912,C0322,W0704
 
 """
 
@@ -14,12 +14,14 @@ Desc: Command-line tool for listing Python packages installed by setuptools,
 
 Author: Rob Cakebread <gentoodev a t gmail.com>
 
-License  : PSF (Python Software Foundation License)
+License : GNU General Public License Version 2 (See COPYING)
 
 """
 
 __docformat__ = 'restructuredtext'
 
+import pprint
+import os
 import sys
 import optparse
 import pkg_resources
@@ -28,8 +30,10 @@ from distutils.sysconfig import get_python_lib
 
 from yolk import __version__
 from yolk.metadata import get_metadata
-from yolk.yolklib import Distributions, get_highest_version
+from yolk.yolklib import get_highest_installed, get_highest_version, Distributions
 from yolk.pypi import CheeseShop
+from yolk.setuptools_support import get_download_uri
+from yolk.plugins import load_plugins
 
 
 class Usage(Exception):
@@ -42,7 +46,7 @@ class Usage(Exception):
         sys.exit(2)
 
 
-#Functions for obtaining info about packages installed with setuptools
+#Functions for obtaining info about packages installed by setuptools
 ##############################################################################
 
 
@@ -59,13 +63,14 @@ def get_pkglist():
 
 def show_updates(package_name="", version=""):
     """Check installed packages for available updates on PyPI"""
+    pypi = CheeseShop()
 
     dists = Distributions()
     for pkg in get_pkglist():
         for (dist, active) in dists.get_distributions("all", pkg,
-                dists.get_highest_installed(pkg)):
+                get_highest_installed(pkg)):
             (project_name, versions) = \
-                    PYPI.query_versions_pypi(dist.project_name, True)
+                    pypi.query_versions_pypi(dist.project_name, True)
             if versions:
 
                 #PyPI returns them in chronological order,
@@ -83,14 +88,39 @@ def show_updates(package_name="", version=""):
                                 newest)
 
 
-def show_distributions(show, project_name, version, show_metadata,
-                       fields):
+def get_plugin(method, options):
+    """Return plugin object if CLI option is activated and method exists"""
+    all_plugins = []
+    for entry_point in pkg_resources.iter_entry_points('yolk.plugins'):
+        plugin_obj = entry_point.load()
+        plugin = plugin_obj()
+        plugin.configure(options, None)
+        if plugin.enabled:
+            if not hasattr(plugin, method):
+                print >> sys.stderr, \
+                        "Error: plugin has no method: %s" % method
+                plugin = None
+            else:
+                all_plugins.append(plugin)
+    return all_plugins
+
+def show_distributions(show, project_name, version, options):
     """Show list of installed activated OR non-activated packages"""
+    show_metadata = options.metadata
+    fields = options.fields
 
-    #When using unionfs in livecd's such as Knoppix we remove prefixes
-    #otherwise all packages show as development
+    #Search for any plugins with active CLI options with add_column() method
+    plugins = get_plugin("add_column", options)
 
+    #Some locations show false positive for 'development' packages:
     ignores = ["/UNIONFS", "/KNOPPIX.IMG"]
+
+    #Check if we're in a workingenv
+    #See http://cheeseshop.python.org/pypi/workingenv.py
+    workingenv = os.environ.get('WORKING_ENV')
+    if workingenv:
+        ignores.append(workingenv)
+
     dists = Distributions()
     results = None
     for (dist, active) in dists.get_distributions(show, project_name,
@@ -99,24 +129,38 @@ def show_distributions(show, project_name, version, show_metadata,
         for prefix in ignores:
             if dist.location.startswith(prefix):
                 dist.location = dist.location.replace(prefix, "")
+        #Case-insensitve search because of Windows
         if dist.location.lower().startswith(get_python_lib().lower()):
             develop = ""
         else:
             develop = dist.location
         if metadata:
-            print_metadata(metadata, develop, active,
-                           show_metadata, fields)
+            add_column_text = ""
+            for my_plugin in plugins:
+                #See if package is 'owned' by a package manager such as portage, apt etc.
+                #add_column_text += my_plugin.add_column(filename) + " "
+                add_column_text += my_plugin.add_column(dist) + " "
+            print_metadata(metadata, develop, active, options, add_column_text)
         else:
             print dist + " has no metadata"
         results = True
-    if show == "all" and results and fields:
+    if not results:
+        if version:
+            pkg_spec = "%s==%s" % (project_name, version)
+        else:
+            pkg_spec = "%s" % project_name
+
+        print >> sys.stderr, "%s is not installed." % pkg_spec
+        return 2
+    elif show == "all" and results and fields:
         print "Versions with '*' are non-active."
         print "Versions with '!' are deployed in development mode."
 
 
-def print_metadata(metadata, develop, active, show_metadata,
-                   fields):
+def print_metadata(metadata, develop, active, options, installed_by):
     """Print out formatted metadata"""
+    show_metadata = options.metadata
+    fields = options.fields
 
     version = metadata['Version']
 
@@ -138,7 +182,7 @@ def print_metadata(metadata, develop, active, show_metadata,
         else:
             development_status = "development (%s)" % develop
     else:
-        development_status = ""
+        development_status = installed_by
     status = "%s %s" % (active_status, development_status)
     if fields:
         print '%s (%s)%s %s' % (metadata['Name'], version, active_status,
@@ -166,6 +210,7 @@ def print_metadata(metadata, develop, active, show_metadata,
                 print '    %s: %s' % (field, metadata[field])
 
 
+
 def show_deps(pkg_ver):
     """Show dependencies for package(s)"""
 
@@ -214,19 +259,28 @@ You can also specify a package name and version:
 
 def show_download_links(package_name, version, file_type):
     """Query PyPI for pkg download URI for a packge"""
+    pypi = CheeseShop()
 
-    for url in PYPI.get_download_urls(package_name, version, file_type):
+    #Try Cheese Shop first
+    url = None
+    for url in pypi.get_download_urls(package_name, version, file_type):
         print url
 
+    #Try setuptools if there is no link on Cheese Shop
+    if not url:
+        url = get_download_uri(file_type, package_name, version)
+        if url:
+            print url
 
 def browse_website(package_name, browser=None):
     """Launch web browser at project's homepage"""
 
+    pypi = CheeseShop()
     #Get verified name from pypi.
 
-    (pypi_project_name, versions) = PYPI.query_versions_pypi(package_name)
+    (pypi_project_name, versions) = pypi.query_versions_pypi(package_name)
     if len(versions):
-        metadata = PYPI.release_data(pypi_project_name, versions[0])
+        metadata = pypi.release_data(pypi_project_name, versions[0])
         if metadata.has_key("home_page"):
             print "Launching browser: %s" % metadata["home_page"]
             if browser == 'konqueror':
@@ -244,27 +298,28 @@ def browse_website(package_name, browser=None):
 
 def show_pkg_metadata_pypi(package_name, version):
     """Show pkg metadata queried from PyPI"""
+    pypi = CheeseShop()
+    (pypi_project_name, versions) = \
+            pypi.query_versions_pypi(package_name, False)
 
-    if version:
-        versions = [version]
+    if versions:
+        for ver in versions:
+            metadata = pypi.release_data(pypi_project_name, ver)
+            if metadata:
+                for key in metadata.keys():
+                    print "%s: %s" % (key, metadata[key])
+        
     else:
-
-        #If they don't specify version, show all.
-
-        (package_name, versions) = PYPI.query_versions_pypi(package_name,
-                None)
-
-    for ver in versions:
-        metadata = PYPI.release_data(package_name, ver)
-        for key in metadata.keys():
-            print "%s: %s" % (key, metadata[key])
+        print >> sys.stderr, "I'm afraid we have no %s at The Cheese Shop. Perhaps a little red Leicester?" % \
+            package_name
+        sys.exit(2)
 
 
 def get_all_versions_pypi(package_name, use_cached_pkglist=False):
     """Fetch list of available versions for a package from The Cheese Shop"""
-
+    pypi = CheeseShop()
     (pypi_project_name, versions) = \
-            PYPI.query_versions_pypi(package_name, use_cached_pkglist)
+            pypi.query_versions_pypi(package_name, use_cached_pkglist)
 
     #pypi_project_name may != package_name; it returns the name with correct case
     #i.e. You give beautifulsoup but PyPI knows it as BeautifulSoup
@@ -272,7 +327,7 @@ def get_all_versions_pypi(package_name, use_cached_pkglist=False):
     if versions:
         print_pkg_versions(pypi_project_name, versions)
     else:
-        print >> sys.stderr, "Nothing found on PyPI for %s" % \
+        print >> sys.stderr, "I'm afraid we have no %s at The Cheese Shop. Perhaps a little red Leicester?" % \
             package_name
         sys.exit(2)
 
@@ -334,10 +389,11 @@ def pypi_search(arg, spec):
     """Search PyPI by metadata keyword
     e.g. yolk -S name=yolk
     """
+    pypi = CheeseShop()
 
     spec.insert(0, arg.strip())
     (spec, operator) = parse_search_spec(spec)
-    for pkg in PYPI.search(spec, operator):
+    for pkg in pypi.search(spec, operator):
         if pkg['summary']:
             summary = pkg['summary'].encode('utf-8')
         else:
@@ -351,18 +407,22 @@ def pypi_search(arg, spec):
 def get_rss_feed():
     """Show last 20 package updates from PyPI RSS feed"""
 
-    rss = PYPI.get_rss()
+    pypi = CheeseShop()
+    rss = pypi.get_rss()
+    items = []
     for pkg in rss.keys():
-        print """%s
-    %s
-""" % (pkg, rss[pkg])
-
+        date = rss[pkg][1][:10]
+        #Show packages grouped by date released
+        if not date in items:
+            items.append(date)
+            print date
+        print "  %s - %s" % (pkg, rss[pkg][0])
 
 #Utility functions
 ##############################################################################
 
 
-def parse_pkg_ver(package_spec):
+def parse_pkg_ver(package_spec, installed):
     """Return tuple with package_name and version from CLI args"""
 
     arg_str = ("").join(package_spec)
@@ -376,16 +436,18 @@ def parse_pkg_ver(package_spec):
         (package_name, version) = arg_str.split("==")
         package_name = package_name.strip()
         version = version.strip()
+    if installed:
+        #Find proper case for package name
+        dists = Distributions()
+        package_name = dists.case_sensitive_name(package_name)
     return (package_name, version)
 
-
 def print_pkg_versions(package_name, versions):
     """Print list of versions available for a package"""
 
     for ver in versions:
         print "%s %s" % (package_name, ver)
 
-
 def validate_pypi_opts(opt_parser):
     """Check for sane parse options"""
 
@@ -400,19 +462,48 @@ Examples:
   PackageName==2.0"""
 
 
+def show_entry_map(dist):
+    """Show entry map for a distribution"""
+    pprinter = pprint.PrettyPrinter()
+    try:
+        pprinter.pprint(pkg_resources.get_entry_map(dist))
+    except pkg_resources.DistributionNotFound:
+        print "Distribution not found: %s" % dist
+
+def show_entry_points(module):
+    """Show entry points for a module"""
+    found = False
+    for entry_point in pkg_resources.iter_entry_points(module):
+        found = True
+        try:
+            plugin = entry_point.load()
+            print plugin.__module__
+            print "   %s" % entry_point
+            if plugin.__doc__:
+                print plugin.__doc__
+            print
+        except ImportError:
+            pass
+    if not found:
+        print "No entry points found for %s" % module
+
 def setup_opt_parser():
     """Setup the optparser"""
 
     usage = "usage: %prog [options]"
     opt_parser = optparse.OptionParser(usage=usage)
 
-    opt_parser.add_option("-v", "--version", action='store_true', dest=
+    opt_parser.add_option("--version", action='store_true', dest=
                           "version", default=False, help=
                           "Show yolk version and exit.")
 
+    opt_parser.add_option("-v", "--verbose", action='store_true', dest=
+                          "verbose", default=False, help=
+                          "Be more verbose.")
+
     group_local = optparse.OptionGroup(opt_parser,
             "Query installed Python packages",
-            "The following options show information about Python packages installed by setuptools. Activated packages are normal packages on sys.path that can be imported. Non-activated packages need 'pkg_resources.require()' before they can be imported, such as packages installed with 'easy_install --multi-version'")
+            "The following options show information about installed Python packages. Activated packages are normal packages on sys.path that can be imported. Non-activated packages need 'pkg_resources.require()' before they can be imported, such as packages installed with 'easy_install --multi-version'. PKG_SPEC can be either a package name or package name and version e.g. Paste==0.9")
 
     group_local.add_option("-l", "--list", action='store_true', dest=
                            "all", default=False, help=
@@ -423,6 +514,16 @@ def setup_opt_parser():
                            'List activated packages installed by ' +
                            'setuptools. Use PKG_SPEC to narrow results.')
 
+    group_local.add_option("--entry-points", action='store',
+                           dest="entry_points", default=False, help=
+                           'List entry points for a module. e.g. --entry-points yolk.plugins',
+                            metavar="MODULE")
+
+    group_local.add_option("--entry-map", action='store',
+                           dest="entry_map", default=False, help=
+                           'List entry map for a distribution. e.g. --entry-map yolk',
+                           metavar="PACKAGE_NAME")
+
     group_local.add_option("-n", "--non-activated", action='store_true',
                            dest="nonactive", default=False, help=
                            'List non-activated packages installed by ' +
@@ -486,11 +587,20 @@ def setup_opt_parser():
                           "listed on PyPI. (Use with PKG_NAME)")
     opt_parser.add_option_group(group_local)
     opt_parser.add_option_group(group_pypi)
+    # add opts from plugins
+    all_plugins = []
+    for plugcls in load_plugins(others=True):
+        plug = plugcls()
+        try:
+            plug.add_options(opt_parser)
+        except AttributeError:
+            pass
+
     return opt_parser
 
 
 def main():
-    """Main function"""
+    """Parse options and perform actions"""
 
     opt_parser = setup_opt_parser()
     (options, remaining_args) = opt_parser.parse_args()
@@ -501,35 +611,48 @@ def main():
         opt_parser.print_help()
         sys.exit(2)
 
+    if options.entry_points:
+        show_entry_points(options.entry_points)
+        sys.exit()
+    if options.entry_map:
+        show_entry_map(options.entry_map)
+        sys.exit()
+
+    #Options that depend on querying installed packages, not PyPI.
+    #We find the proper case for package names if they are installed,
+    #otherwise PyPI returns the correct case.
+    if options.depends or options.all or options.active or options.nonactive:
+        want_installed = True
+    else:
+        want_installed = False
     if remaining_args:
-        (package, version) = parse_pkg_ver(remaining_args)
+        (package, version) = parse_pkg_ver(remaining_args, want_installed)
+        if want_installed and not package:
+            print >> sys.stderr, "%s is not installed." % remaining_args[0]
+            sys.exit(2)
     else:
         package = version = None
+
     if options.search:
         pypi_search(options.search, remaining_args)
     elif options.version:
-
         print "Version %s" % __version__.VERSION
     elif options.depends:
         show_deps(remaining_args)
     elif options.all:
         if options.active or options.nonactive:
             opt_parser.error("Choose either -l, -n or -a, not combinations of those.")
-        show_distributions("all", package, version, options.metadata,
-                           options.fields)
+        show_distributions("all", package, version, options)
     elif options.active:
         if options.all or options.nonactive:
             opt_parser.error("Choose either -l, -n or -a, not combinations of those.")
 
-        show_distributions("active", package, version, options.metadata,
-                           options.fields)
+        show_distributions("active", package, version, options)
     elif options.nonactive:
         if options.active or options.all:
             opt_parser.error("Choose either -l, -n or -a, not combinations of those.")
-        show_distributions("nonactive", package, version, options.metadata,
-                           options.fields)
+        show_distributions("nonactive", package, version, options)
     elif options.versions_available:
-
         get_all_versions_pypi(package, False)
     elif options.browse_website:
         browse_website(package)
@@ -543,11 +666,8 @@ def main():
         show_pkg_metadata_pypi(package, version)
     else:
         opt_parser.print_help()
-        sys.exit(2)
-
-
-PYPI = CheeseShop()
+        return 2
 
 if __name__ == "__main__":
-    main()
+    sys.exit(main())
 
