@@ -1,6 +1,7 @@
 #!/usr/bin/python
 # -*- coding: utf-8 -*-
-# pylint: disable-msg=C0301,W0613,W0612
+# pylint: disable-msg=C0301,W0613,W0612,W0212
+# Fix W0212 Access to a protected member _headers of a client class
 
 """
 
@@ -9,7 +10,7 @@ metadata.py
 
 Author   : Rob Cakebread <gentoodev@gmail.com>
 
-License  : PSF (Python Software Foundation License)
+License  : GNU General Public License Version 2
 
 Desc     : Return metadata for Python distribution installed by setuptools
            in a dict
@@ -28,8 +29,8 @@ def get_metadata(dist):
     if not dist.has_metadata('PKG-INFO'):
         return
 
-    md = dist.get_metadata('PKG-INFO')
-    msg = email.message_from_string(md) 
+    my_metadata = dist.get_metadata('PKG-INFO')
+    msg = email.message_from_string(my_metadata) 
     metadata = {}
     for header in [l for l in msg._headers]:
         metadata[header[0]] = header[1]
